#include "sys.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>

#ifndef WINDOWS
	#include <errno.h>
	#include <sys/types.h>
	#include <unistd.h>
	#include <time.h>
	#include <langinfo.h>
#endif

//[of]:newString
static String newString(String src)
{
	#ifdef WINDOWS
		size_t len = wcslen(src) + 1;
		String str = (String)csys_allocate(len * sizeof(CodeUnit));
		memcpy(str, src, len * sizeof(CodeUnit));
		return str;
	#else
		size_t len = strlen(src) + 1;
		String str = (String)csys_allocate(len * sizeof(CodeUnit));
		memcpy(str, src, len * sizeof(CodeUnit));
		return str;
	#endif
}
//[cf]
//[of]:newSystemErrorMessage
//[c]Returns the message corresponding to the error returned by an ERR_SYSTEM
//[c]
ENTRY String csys_newSystemErrorMessage()
{
	String str;

	#ifdef WINDOWS
		size_t len;
		wchar_t* msg;
		DWORD error = GetLastError();
		LPWSTR buf;
		DWORD res = FormatMessage(	FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, 
			NULL,
			error, 
			LANG_USER_DEFAULT, 
			(LPTSTR)&buf, 
			4096,
			NULL);
		if( res != 0 )
		{
			str = newString(buf);
			LocalFree(buf);
		}
		else
		{
			str = newString(L"");
		}	
	#else
		String tmp;
		int res = csys_newStringFromLocale(&tmp, strerror(errno));
		if( res == OK )
		{
			str = newString(tmp);
			csys_deleteLocale(tmp);
		}
		else
		{
			str = newString("");
		}
	#endif	

	return str;
}
//[cf]

#ifdef WINDOWS
//[of]:	newStringFromWide
//[c]Converts the UTF-16 encoded input string to an string.
//[c]
//[c]PARAMETERS
//[c]	out	--	A pointer to store the resulting string.
//[c]	input	--	The input string encoded in UTF-16.
//[c]
//[c]RETURN VALUE
//[c]	An error code, if not OK, the output string is not set and must not be 
//[c]	deleted. Otherwise it must be deleted with deleteString.
//[c]
ENTRY SysErrorCode csys_newStringFromWide(String* out, const wchar* input)
{
	// On windows, String is encoded in UTF-16, so no conversion is
	// required
	*out = (String)input;
	return OK;
}
//[cf]
//[of]:	deleteString
//[c]Deletes a string created with newStringFromWide.
//[c]
ENTRY void csys_deleteString(String string)
{
	// Nothing to do
}
//[cf]
//[of]:	newWideFromString
ENTRY SysErrorCode csys_newWideFromString(wchar**out, String input)
{
	*out = (wchar*)input;
	return OK;
}
//[cf]
//[of]:	deleteWide
//[c]Deletes a string created with newWideFromString.
//[c]
ENTRY void csys_deleteWide(const wchar* string)
{
	// Nothing to do
}
//[cf]
#else
	static int localeInitialized = 0;
	static int localeIsUtf8 = 0;

//[of]:	Converter
typedef struct 
{
	iconv_t	cd;
	size_t	outlen;
	size_t	inlen;
	char*	input;
	char*	inbuf;
	size_t	insize;
	char*	outbuf;
	size_t	outsize;
	
} Converter;

#define bufSize 1024

static SysErrorCode cv_open(Converter* cv, char* input, const char* tocode, const char* fromcode)
{
	char buf[bufSize];
	int res;

	cv->cd = iconv_open(tocode, fromcode);
	if( cv->cd == NULL )
		return ERR_SYSTEM;
	
	cv->input = input;
	cv->inlen = strlen(input);
	cv->outlen = 0;
	cv->inbuf = input;
	cv->insize = cv->inlen;
	
	// Determine the length we need.
	while( cv->insize > 0 )
	{
		cv->outbuf = buf;
		cv->outsize = bufSize;
		res = iconv(cv->cd, &cv->inbuf, &cv->insize, &cv->outbuf, &cv->outsize);
		if( res == -1 && errno != E2BIG )
		{
			// An error could erase the original error
			iconv_close(cv->cd); 
			return ERR_SYSTEM;
		}
		cv->outlen += cv->outbuf - buf;
	}
	cv->outbuf = buf;
	cv->outsize = bufSize;
	res = iconv(cv->cd, NULL, NULL, &cv->outbuf, &cv->outsize);
	if( res == -1 )
	{
		// An error could erase the original error
		iconv_close(cv->cd); 
		return ERR_SYSTEM;
	}
	cv->outlen += cv->outbuf - buf;
	return OK;
}

static void cv_convert(Converter* cv, char* output)
{
	// Return to the initial state
	iconv(cv->cd, NULL, NULL, NULL, NULL);

	// Do the conversion for real.
	cv->inbuf = cv->input;
	cv->insize = cv->inlen;
	cv->outbuf = output;
	cv->outsize = cv->outlen;
	iconv(cv->cd, &cv->inbuf, &cv->insize, &cv->outbuf, &cv->outsize);
	
	iconv_close(cv->cd);
}
//[cf]
//[of]:	initializeLocale
void initializeLocale()
{
	if( !localeInitialized )
	{
		// Thread safety: this part is not thread safe because setlocale is not
		// thread safe. setlocale should be invoked at the beginning of the 
		// application or before creating any additional thread
		const char* codeset;
		setlocale(LC_ALL, "");
		codeset = nl_langinfo(CODESET);
		localeIsUtf8 = strcasecmp(codeset, "UTF-8") == 0;
		localeInitialized = 1;
	}
}
//[cf]
//[of]:	needConversion
static int needConversion()
{
	initializeLocale();
	return !localeIsUtf8;
}
//[cf]
//[of]:	convert
//[c]Converts the encoding of a string..
//[c]
//[c]PARAMETERS
//[c]	out	--	A pointer to store the resulting string.
//[c]	input	--	The input string.
//[c]
//[c]RETURN VALUE
//[c]	An error code, if not OK, the output string is not set and must not be 
//[c]	deleted.
//[c]
static SysErrorCode convert(char** out, char* input, const char* tocode, const char* fromcode)
{
	Converter cv;
	char* output;
	SysErrorCode res = cv_open(&cv, input, tocode, fromcode);
	if( res != OK )
		return res;

	// At this point, the input has been successfully converted, 
	// we assume that iconv cannot fail, return code are ignored.
	//
	
	// Allocate the buffer
	output = (char*)csys_allocate(cv.outlen+1);
	
	cv_convert(&cv, output);
	output [cv.outlen] = 0;
	*out = output;
	
	return OK;
}
//[cf]
//[of]:	validateUTF8
// Type for UTF-8
// 0 - ASCII
// 1 - 1 continuation
// 2 - 2 continuations
// 3 - 3 continuations
// 4 - invalid UTF-8
// 5 - continuation
static char utf8Type[256] = {
	6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, // 80-8F
	5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, // 90-9F
	5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, // A0-AF
	5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, // B0-BF
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4};

static int validateUTF8(const char* input)
{
	const unsigned char* p = (const unsigned char*)input;
	int n = 0;
	for(;;)
	{
		switch( utf8Type [*p++] )
		{
		case 0: // ASCII
			if( n != 0 )
				return 0;
			break;
		case 1:
			if( n != 0 )
				return 0;
			n = 1;
			break;
		case 2:
			if( n != 0 )
				return 0;
			n = 2;
			break;
		case 3:
			if( n != 0 )
				return 0;
			n = 3;
			break;
		case 4:
			return 0;
		case 5:
			if( n == 0 )
				return 0;
			n --;
			break;
		case 6:
			return n == 0;
		}
	}
}
//[cf]
//[of]:	newStringFromLocale
//[c]Converts the input string encoded with current locale to an UTF-8 string.
//[c]
//[c]PARAMETERS
//[c]	out	--	A pointer to store the resulting string.
//[c]	input	--	The input string encoded with the current locale.
//[c]
//[c]RETURN VALUE
//[c]	An error code, if not OK, the output string is not set and must not be 
//[c]	deleted. Otherwise it must be deleted with deleteString.
//[c]
ENTRY SysErrorCode csys_newStringFromLocale(String* out, const char* input)
{
	if( needConversion() )
	{
		return convert(out, (char*)input, "UTF-8", "");
	}
	else
	{
		if( !validateUTF8(input) )
			return ERR_INVALID_UTF8;
		*out = (String)input;
		return OK;
	}
}
//[cf]
//[of]:	newCStringFromLocale
//[c]Same as csys_newCStringFromLocale, but the string is always allocated (it is 
//[c]a String that will be deleted from Copper, not a SysString intended to be 
//[c]deleted by csys_deleteString).
//[c]
ENTRY SysErrorCode csys_newCStringFromLocale(String* out, const char* input)
{
	// Ensure that locale is initialized before calling convert
	initializeLocale();
	
	// Always convert to force allocation of a new string
	return convert(out, (char*)input, "UTF-8", "");
}
//[cf]
//[of]:	deleteString
//[c]Deletes a string created with newStringFromLocale.
//[c]
ENTRY void csys_deleteString(String string)
{
	if( needConversion() )
	{
		int len = strlen(string);
		csys_free((void*)string, len + 1);
	}
}
//[cf]
//[of]:	newLocaleFromString
ENTRY SysErrorCode csys_newLocaleFromString(char**out, String input)
{
	if( needConversion() )
	{
		return convert(out, (char*)input, "", "UTF-8");
	}
	else
	{
		*out = (char*)input;
		return OK;
	}
}
//[cf]
//[of]:	deleteLocale
//[c]Deletes a string created with newLocaleFromString.
//[c]
ENTRY void csys_deleteLocale(const char* string)
{
	// Same as deleteString
	csys_deleteString((String)string);
}
//[cf]
#endif

//[of]:StringBuffer
void StringBuffer_initialize(StringBuffer* sb)
{
	sb->base = sb->local;
	sb->size = 0;
	sb->allocated = STRING_BUFFER_LOCAL_SIZE;
}

void StringBuffer_release(StringBuffer* sb)
{
	if( sb->base != sb->local )
		csys_free(sb->base, sb->allocated);
}

#ifdef WINDOWS
	SysErrorCode StringBuffer_addWide(StringBuffer* sb, const wchar* str)
	{
		// On windows, the StringBuffer is encoded as UTF-16 as well as
		// the windows strings, just add
		StringBuffer_addString(sb, (String)str);
		return OK;
	}
#else
	SysErrorCode StringBuffer_addLocale(StringBuffer* sb, const char* str)
	{
		SysErrorCode res;
		char* output;
		
		if( needConversion() )
		{
			Converter cv;
			int res = cv_open(&cv, (char*)str, "UTF-8", "");
			if( res != OK )
				return res;
	
			output = StringBuffer_reserve(sb, cv.outlen);
			cv_convert(&cv, output);
			sb->size += cv.outlen;
		}
		else
		{
			if( !validateUTF8(str) )
				return ERR_INVALID_UTF8;
	
			size_t len = strlen(str);
			output = StringBuffer_reserve(sb, len);
			memcpy(output, str, len);
			sb->size += len;
		}
		return OK;
	}
#endif

void StringBuffer_addString(StringBuffer* sb, const String str)
{
	#ifdef WINDOWS
		size_t len = wcslen(str);
	#else
		size_t len = strlen(str);
	#endif
	String out = StringBuffer_reserve(sb, len);
	memcpy(out, str, len * sizeof(CodeUnit));
	sb->size += len;
}

String StringBuffer_newString(StringBuffer* sb)
{
	String buf = (String)csys_allocate((sb->size+1) * sizeof(CodeUnit));
	memcpy(buf, sb->base, sb->size * sizeof(CodeUnit));
	buf[sb->size] = 0;
	return buf;
}

String StringBuffer_reserve(StringBuffer* sb, size_t len)
{
	if( sb->size + len > sb->allocated )
	{
		// Compute new capacity
		size_t newCapacity = (sb->size + len) * 3 / 2;
		
		// Allocate new buffer
		String q = (String)csys_allocate(newCapacity * sizeof(CodeUnit));
		
		// Copy the old buffer into the new one
		memcpy(q, sb->base, sb->size * sizeof(CodeUnit));
		
		// Free the old buffer
		if( sb->base != sb->local )
			csys_free(sb->base, sb->allocated * sizeof(CodeUnit));
		
		// Update object
		sb->allocated = newCapacity;
		sb->base = q;
	}
	return sb->base + sb->size;
}
//[cf]
//[of]:MemoryBuffer
void MemoryBuffer_initialize(MemoryBuffer* mb)
{
	mb->base = NULL;
	mb->size = 0;
	mb->allocated = 0;
}

void MemoryBuffer_release(MemoryBuffer* mb)
{
	csys_free(mb->base, mb->allocated);
}

void MemoryBuffer_add(MemoryBuffer* mb, const char* data, size_t size)
{
	char* out = MemoryBuffer_reserve(mb, size);
	memcpy(out, data, size);
	mb->size += (int)size;
}

char* MemoryBuffer_reserve(MemoryBuffer* mb, size_t len)
{
	if( mb->size + len > mb->allocated )
	{
		char* q;
		
		// Compute new capacity
		size_t newCapacity = (mb->size + len) * 3 / 2;
		if( newCapacity < 4096 )
			newCapacity = 4096;
		
		// Allocate new buffer
		q = (char*)csys_allocate(newCapacity);
		
		// Copy the old buffer into the new one
		memcpy(q, mb->base, mb->size);
		
		// Free the old buffer
		csys_free(mb->base, mb->allocated);
		
		// Update object
		mb->allocated = newCapacity;
		mb->base = q;
	}
	return mb->base + mb->size;
}
//[cf]
//[of]:TextDecoder
//[of]:open
SysErrorCode csys_TextDecoder_open(const char* fromCode, TextDecoder** output)
{
	#ifdef WINDOWS
		TextDecoder* d = (TextDecoder*)csys_allocate(sizeof(TextDecoder));
		
		if( stricmp(fromCode, "utf-8") == 0 )
		{
			d->fromEncoding = CE_OTHER;
			d->fromCodePage = CP_UTF8;
		}
		else if( stricmp(fromCode, "utf-16le") == 0 )
		{
			d->fromEncoding = CE_UTF16LE;
		}
		else if( stricmp(fromCode, "utf-16be") == 0 )
		{
			d->fromEncoding = CE_UTF16BE;
		}
		else if( stricmp(fromCode, "iso88591-1") == 0 )
		{
			d->fromEncoding = CE_OTHER;
			d->fromCodePage = 28591;
		}
		else
		{
			d->fromEncoding = CE_OTHER;
			d->fromCodePage = CP_ACP;
		}
		MemoryBuffer_initialize(&d->buffer);
	#else
		iconv_t cd = iconv_open("utf-8", fromCode);
		if( cd == (iconv_t)-1 )
			return ERR_SYSTEM;
		
		TextDecoder* d = (TextDecoder*)csys_allocate(sizeof(TextDecoder));
		d->cd = cd;
		d->minibufSize = 0;
	#endif
	*output = d;
	return OK;
}
//[cf]
//[of]:convert
//[of]:decode
#ifdef WINDOWS
	static int decode(TextDecoder* decoder, StringBuffer* output, const char* input, size_t size)
	{
		switch(decoder->fromEncoding)
		{
			case CE_UTF16LE:
			{
				const char* s;
				const char* limit;
				CodeUnit c;
				int n;
				String p;

				// The number of bytes must be even
				if( (size & 1) == 1 )
					return 0;
				
				// The last code unit must not be a leading surrogate
				c = (input[size - 1] << 8) + input[size - 2];
				if( (c & 0xFC00) == 0xDC00 )
					return 0;
				
				// Validate the sequence
				n = size / sizeof(CodeUnit);
				p = StringBuffer_reserve(output, n);
				s = input;
				limit = input + size;
				while( s != limit ) 
				{
					c = s[0] + (s[1] << 8);
					*p++ = c;
					s += sizeof(CodeUnit);
				}
				output->size += n;
				return n;
			}
			case CE_UTF16BE:
			{
				const char* s;
				const char* limit;
				CodeUnit c;
				int n;
				String p;

				// The number of bytes must be even
				if( (size & 1) == 1 )
					return 0;
				
				// The last code unit must not be a leading surrogate
				c = (input [size - 2] << 8) + input [size - 1];
				if( (c & 0xFC00) == 0xDC00 )
					return 0;
				
				// Validate the sequence
				n = size / sizeof(CodeUnit);
				p = StringBuffer_reserve(output, n);
				s = input;
				limit = s + size;
				while( s != limit ) 
				{
					c = s[1] + (s[0] << 8);
					*p++ = c;
					s += sizeof(CodeUnit);
				}
				output->size += n;
				return n;
			}
			default: // case CE_OTHER:
			{
				String str;
				int n = MultiByteToWideChar(decoder->fromCodePage, 0, (LPCSTR)input, (int)size, NULL, 0);
				if( n != 0 )
				{
					String str = StringBuffer_reserve(output, (size_t)n);
					MultiByteToWideChar (decoder->fromCodePage, 0, (LPCSTR)input, (int)size, (LPWSTR)str, n);
					output->size += n;
				}
				return n;
			}
		}
	}
#endif
//[cf]

SysErrorCode csys_TextDecoder_convert(TextDecoder* decoder, StringBuffer* output, const char* input, size_t size)
{
	int originalOutputSize;
	
	// Nothing to convert
	if( size == 0 )
		return OK;
		
	// Save the original size of the output
 	originalOutputSize = output->size;

	#ifdef WINDOWS
		if( decoder->buffer.size == 0 )
		{
			// Nothing in the buffer: attempt to convert the current data
			if( decode(decoder, output, input, size) == 0 )
				// The conversion has failed: append the input to the buffer and return
				MemoryBuffer_add(&decoder->buffer, input, size);
			// Always return OK (errors may be temporaries because of incomplete data)
		}
		else
		{
			// There is something in the buffer: add the intput to the buffer 
			MemoryBuffer_add(&decoder->buffer, input, size);
			// And try conversion
			if( decode(decoder, output, input, size) != 0 )
				// The conversion has succeeded: reset the buffer
				decoder->buffer.size = 0;
			// Always return OK (errors may be temporaries because of incomplete data)
		}
	#else
		size_t wr;
		char* inbuf;
		size_t insize;
		char* outbuf;
		size_t outsize;
		
		// Attempt to flush the incomplete sequence
		if( decoder->minibufSize != 0 )
		{
			outbuf = (char*)StringBuffer_reserve(output, MAX_SEQUENCE_SIZE);
			while( decoder->minibufSize != MAX_SEQUENCE_SIZE && size != 0 )
			{
				// Add one byte
				decoder->minibuf[decoder->minibufSize++] = *input++;
				size--;
				
				// Try conversion
				inbuf = decoder->minibuf;
				insize = decoder->minibufSize;
				outsize = MAX_SEQUENCE_SIZE * sizeof(CodeUnit);
				wr = iconv(decoder->cd, &inbuf, &insize, &outbuf, &outsize);
				if( wr != -1 )
					break;
				
				// Continue as long as we get an incomplete multibyte sequence
				if( errno != EINVAL )
					return ERR_SYSTEM;
					
				// Exit if no more bytes
				if( size == 0 )
					return OK;
			}
			output->size = (CodeUnit*)outbuf - output->base;
			decoder->minibufSize = 0;
		}
		
		// Initialize the input buffer
		inbuf = (char*)input;
		insize = size;
		
		for(;;)
		{
			// Convert it 
			size_t outBufferSize = size + MAX_SEQUENCE_SIZE; // Good estimate
			char* outbuf = (char*)StringBuffer_reserve(output, outBufferSize);
			size_t outsize = outBufferSize * sizeof(CodeUnit);
			size_t wr = iconv(decoder->cd, &inbuf, &insize, &outbuf, &outsize);
			output->size = (CodeUnit*)outbuf - output->base;
			
			if( wr == -1 )
			{
				// 3. An incomplete multibyte sequence is left
				if( errno == EINVAL )
				{
					// Safety check
					if( insize == 0 || insize > MAX_SEQUENCE_SIZE )
						return ERR_SYSTEM;
					
					// Move it to the minibuf
					memcpy(decoder->minibuf, inbuf, insize);
					decoder->minibufSize = insize;
					
					// And return ok
					break;
				}
				// 4. The output buffer is full
				else if( errno == E2BIG )
				{
					// Repeat conversion
				}
				// 1. invalid sequence: return an error
				else
					return ERR_SYSTEM;
			}
			else
			{
				// 2. The input byte sequence has been entirely converted.
				break;
			}
		}
	#endif
	
	// Test if the converted result contains a nul code-unit.
	// It is not valid to have a zero in the text, it would break
	// the computation of string size.
	//
	// A StringBuffer can theoritically handle it, but it could not
	// be converted correctly to a string.
	//
	{
		String s = output->base + originalOutputSize;
		String limit = output->base + output->size;
		while( s != limit )
		{
			if( *s == 0 )
			{
				// Truncate the output
				output->size = s - output->base;
				return ERR_NUL_CODE_UNIT;
			}
			s++;
		}
	}
	
	return OK;
}
//[cf]
//[of]:close
SysErrorCode csys_TextDecoder_close(TextDecoder* decoder)
{
	#ifdef WINDOWS
		size_t size = decoder->buffer.size;
		
		MemoryBuffer_release(&decoder->buffer);
		csys_free((void*)decoder, sizeof(TextDecoder));

		// There are still unconsumed bytes: incomplete sequence
		if( size != 0 )
		{
			SetLastError(ERROR_NO_UNICODE_TRANSLATION);
			return ERR_SYSTEM;
		}
		
		return OK;
	#else
		iconv_t cd = decoder->cd;
		size_t minibufSize = decoder->minibufSize;

		csys_free((void*)decoder, sizeof(TextDecoder));
		
		int res = iconv_close(cd);
		if( res != 0 )
			return ERR_SYSTEM;
			
		// There are still unconsumed bytes: incomplete sequence
		if( minibufSize != 0 )
		{
			errno = EILSEQ;
			return ERR_SYSTEM;
		}

		return OK;
	#endif
}
//[cf]
//[cf]
//[of]:TextEncoder
//[of]:open
SysErrorCode csys_TextEncoder_open(const char* toCode, TextEncoder** output)
{
	#ifdef WINDOWS
		TextEncoder* e = (TextEncoder*)csys_allocate(sizeof(TextEncoder));
		
		if( stricmp(toCode, "utf-8") == 0 )
		{
			e->toEncoding = CE_OTHER;
			e->toCodePage = CP_UTF8;
		}
		else if( stricmp(toCode, "utf-16le") == 0 )
		{
			e->toEncoding = CE_UTF16LE;
		}
		else if( stricmp(toCode, "utf-16be") == 0 )
		{
			e->toEncoding = CE_UTF16BE;
		}
		else if( stricmp(toCode, "iso88591-1") == 0 )
		{
			e->toEncoding = CE_OTHER;
			e->toCodePage = 28591;
		}
		else
		{
			e->toEncoding = CE_OTHER;
			e->toCodePage = CP_ACP;
		}
	#else
		iconv_t cd = iconv_open(toCode, "utf-8");
		if( cd == (iconv_t)-1 )
			return ERR_SYSTEM;
		
		TextEncoder* e = (TextEncoder*)csys_allocate(sizeof(TextEncoder));
		e->cd = cd;
	#endif

	*output = e;
	return OK;
}
//[cf]
//[of]:convert
SysErrorCode csys_TextEncoder_convert(TextEncoder* encoder, MemoryBuffer* output, String input, int size)
{
	// Nothing to convert
	if( size == 0 )
		return OK;
		
	#ifdef WINDOWS
		switch( encoder->toEncoding )
		{
			case CE_UTF16LE:
			{
				int n = size * sizeof(CodeUnit);
				char* d = MemoryBuffer_reserve(output, n);
				memcpy(d, input, n);
				output->size += n;
				break;
			}
			case CE_UTF16BE:
			{
				int n = size * sizeof(CodeUnit);
				const char* s = (const char*)input;
				const char* limit = s + n;
				char* d = MemoryBuffer_reserve(output, n);
				while( s != limit )
				{
					d[0] = s[1];
					d[1] = s[0];
					d += sizeof(CodeUnit);
					s += sizeof(CodeUnit);
				}
				output->size += n;
				break;
			}
			default: // case CE_OTHER:
			{
				char* d;
				int n = WideCharToMultiByte(encoder->toCodePage, 0, (LPCWSTR)input, size, NULL, 0, NULL, NULL);
				if( n == 0 )
					return ERR_SYSTEM;
				d = MemoryBuffer_reserve(output, (size_t)n);
				WideCharToMultiByte(encoder->toCodePage, 0, (LPCWSTR)input, size, (LPSTR)d, n, NULL, NULL);
				output->size += n;
				break;
			}
		}
	#else
		// Initialize the input buffer
		char* inbuf = (char*)input;
		size_t insize = size * sizeof(CodeUnit);;
		
		for(;;)
		{
			// Good estimate (unless UTF-16 or UTF-32)
			// Add the max sequence size in order to prevent an infinite loop
			size_t outBufferSize = size + MAX_SEQUENCE_SIZE;
			char* outbuf = MemoryBuffer_reserve(output, outBufferSize);
			size_t outsize = outBufferSize;
			size_t wr = iconv(encoder->cd, &inbuf, &insize, &outbuf, &outsize);
			output->size = outbuf - output->base;

			// 2. The input byte sequence has been entirely converted.
			if( wr != -1 )
				break;
			// 1. invalid sequence: return an error
			// 3. An incomplete multibyte sequence is left
			if( errno != E2BIG )
				return ERR_SYSTEM;
			// 4. The output buffer is full: repeat conversion
		}
	#endif

	return OK;
}
//[cf]
//[of]:close
SysErrorCode csys_TextEncoder_close(TextEncoder* encoder)
{
	#ifdef WINDOWS
		csys_free((void*)encoder, sizeof(TextEncoder));
		return OK;
	#else
		int res = iconv_close(encoder->cd);
		csys_free((void*)encoder, sizeof(TextEncoder));
		return ( res != 0 ) ? ERR_SYSTEM : OK;
	#endif
}
//[cf]
//[cf]
